(** OCaml bindings for the Xlib library. *)
(*  Copyright (C) 2008 by Florent Monnier <fmonnier@linux-nantes.org>
 *
 * OCaml-Xlib is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 *
 * OCaml-Xlib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OCaml-Xlib.  If not, see:
 * <http://www.gnu.org/licenses/>
 *)

(**
{ul
  {- {{:http://www.x.org/docs/X11/xlib.pdf}Xlib manual (pdf)}}
}
*)

(** {3 Types} *)

type event_mask =
  | KeyPressMask
  | KeyReleaseMask
  | ButtonPressMask
  | ButtonReleaseMask
  | EnterWindowMask
  | LeaveWindowMask
  | PointerMotionMask
  | PointerMotionHintMask
  | Button1MotionMask
  | Button2MotionMask
  | Button3MotionMask
  | Button4MotionMask
  | Button5MotionMask
  | ButtonMotionMask
  | KeymapStateMask
  | ExposureMask
  | VisibilityChangeMask
  | StructureNotifyMask
  | ResizeRedirectMask
  | SubstructureNotifyMask
  | SubstructureRedirectMask
  | FocusChangeMask
  | PropertyChangeMask
  | ColormapChangeMask
  | OwnerGrabButtonMask

type uint = int
type 'a drawable
type _window type window = _window drawable
type _pixmap type pixmap = _pixmap drawable
(** both windows and pixmaps can be used as drawable *)
type gc
type colormap

(** {3 Display} *)

type display
external xOpenDisplay: name:string -> display = "ml_XOpenDisplay"
external xCloseDisplay: dpy:display -> unit = "ml_XCloseDisplay"
external xFlush: dpy:display -> unit = "ml_XFlush"
external xSync: dpy:display -> discard:bool -> unit = "ml_XSync"
external xConnectionNumber: dpy:display -> int = "ml_XConnectionNumber"

external xProtocolVersion: dpy:display -> int = "ml_XProtocolVersion"
external xProtocolRevision: dpy:display -> int = "ml_XProtocolRevision"
external xVendorRelease: dpy:display -> int = "ml_XVendorRelease"
external xServerVendor: dpy:display -> string = "ml_XServerVendor"

external xlibSpecificationRelease: unit -> int = "ml_XlibSpecificationRelease"

(* Server Grabbing *)
external xGrabServer: dpy:display -> unit = "ml_XGrabServer"
external xUngrabServer: dpy:display -> unit = "ml_XUngrabServer"


(** {3 Screen number} *)

type screen_number
external xDefaultScreen: dpy:display -> screen_number = "ml_XDefaultScreen"

external xDisplayWidth: dpy:display -> screen_number -> int = "ml_XDisplayWidth"
external xDisplayHeight: dpy:display -> screen_number -> int = "ml_XDisplayHeight"

external xDefaultDepth: dpy:display -> screen_number -> int = "ml_XDefaultDepth"
external xListDepths: dpy:display -> screen_number -> int array = "ml_XListDepths"

external xScreenCount: dpy:display -> int = "ml_XScreenCount"

#if defined(MLI)
val xScreenNumbers : dpy:display -> screen_number array
(** *)
#else
let xScreenNumbers ~dpy =
  let n = xScreenCount ~dpy in
  Array.init n (fun i -> (Obj.magic i : screen_number))
;;
#endif


(** {3 Pixel Colors} *)

type pixel_color
external xBlackPixel: dpy:display -> screen_number -> pixel_color = "ml_XBlackPixel"
external xWhitePixel: dpy:display -> screen_number -> pixel_color = "ml_XWhitePixel"


(** {3 XColor} *)

type xColor
external new_xColor: unit -> xColor = "ml_alloc_XColor"
(** this type is garbage collected *)

external xAllocNamedColor: dpy:display -> colormap:colormap -> color_name:string -> xColor * xColor = "ml_XAllocNamedColor"
(** the returned values are garbage collected *)

external xColor_set_red: xColor -> int -> unit = "ml_XColor_set_red"
external xColor_set_green: xColor -> int -> unit = "ml_XColor_set_green"
external xColor_set_blue: xColor -> int -> unit = "ml_XColor_set_blue"

type color_flags =
  | DoRed
  | DoGreen
  | DoBlue

external xColor_set_flags: xColor -> color_flags list -> unit = "ml_XColor_set_flags"

external xAllocColor: dpy:display -> colormap:colormap -> xColor -> unit = "ml_XAllocColor"

external xColor_pixel: xColor -> pixel_color = "ml_XColor_pixel"


(** {3 Visual} *)

type visual
type visualID
external xDefaultVisual: dpy:display -> screen_number -> visual = "ml_XDefaultVisual"

external visual_red_mask: visual:visual -> int = "ml_Visual_red_mask"
external visual_green_mask: visual:visual -> int = "ml_Visual_green_mask"
external visual_blue_mask: visual:visual -> int = "ml_Visual_blue_mask"

external visual_bits_per_rgb: visual:visual -> int = "ml_Visual_bits_per_rgb"
external visual_visualid: visual:visual -> visualID = "ml_Visual_visualid"
external visual_map_entries: visual:visual -> int = "ml_Visual_map_entries"
(** color map entries *)


(** {3 xVisualInfo} *)

type xVisualInfo  (* pointer to a structure *)
type xVisualInfo_contents = { (* contents of this structure *)
    visual: visual;
    visual_id: visualID;
    screen_number: screen_number;
    depth: int;
    red_mask: uint;
    green_mask: uint;
    blue_mask: uint;
    colormap_size: int;
    bits_per_rgb: int;
  }

external xVisualInfo_datas: visual_info:xVisualInfo  -> xVisualInfo_contents = "ml_XVisualInfo_contents"
external xFree_xVisualInfo: xVisualInfo -> unit = "ml_XFree_XVisualInfo"


(** {3 Colormap} *)

external xDefaultColormap: dpy:display -> screen_number -> colormap = "ml_XDefaultColormap"
(** do not free the default colormap *)

external xDisplayCells: dpy:display -> screen_number -> int = "ml_XDisplayCells"

type colormap_alloc =
  | AllocNone
  | AllocAll

external xCreateColormap: dpy:display -> win:window -> visual:visual -> alloc:colormap_alloc -> colormap = "ml_XCreateColormap"
external xFreeColormap: dpy:display -> colormap:colormap -> unit = "ml_XFreeColormap"
external xCopyColormapAndFree: dpy:display -> colormap:colormap -> colormap = "ml_XCopyColormapAndFree"
external xSetWindowColormap: dpy:display -> win:window -> colormap:colormap -> unit = "ml_XSetWindowColormap"
(*
XChangeWindowAttributes
*)

(** {4 Managing Installed Colormaps} *)

external xInstallColormap: dpy:display -> colormap:colormap -> unit = "ml_XInstallColormap"
external xUninstallColormap: dpy:display -> colormap:colormap -> unit = "ml_XUninstallColormap"
external xListInstalledColormaps: dpy:display -> win:window -> colormap array = "ml_XListInstalledColormaps"



(** {4 xSetWindowAttributes} *)

type xSetWindowAttributes
external new_xSetWindowAttributes: unit -> xSetWindowAttributes = "ml_XSetWindowAttributes_alloc"
(** the returned [xSetWindowAttributes] is garbage collected *)

type event_mask_list = event_mask list

type cursor

(* {{{ setting xSetWindowAttributes fields *)

(* generated with `cpp winAttr_set.ml.cpp` *)

external winAttr_set_background_pixmap: xSetWindowAttributes -> background_pixmap:pixmap -> unit = "ml_xSetWindowAttributes_set_background_pixmap"
external winAttr_set_background_pixel: xSetWindowAttributes -> background_pixel:pixel_color -> unit = "ml_xSetWindowAttributes_set_background_pixel"
external winAttr_set_border_pixmap: xSetWindowAttributes -> border_pixmap:pixmap -> unit = "ml_xSetWindowAttributes_set_border_pixmap"
external winAttr_set_border_pixel: xSetWindowAttributes -> border_pixel:pixel_color -> unit = "ml_xSetWindowAttributes_set_border_pixel"
external winAttr_set_bit_gravity: xSetWindowAttributes -> bit_gravity:int -> unit = "ml_xSetWindowAttributes_set_bit_gravity"
external winAttr_set_win_gravity: xSetWindowAttributes -> win_gravity:int -> unit = "ml_xSetWindowAttributes_set_win_gravity"
external winAttr_set_backing_store: xSetWindowAttributes -> backing_store:int -> unit = "ml_xSetWindowAttributes_set_backing_store"
external winAttr_set_backing_planes: xSetWindowAttributes -> backing_planes:uint -> unit = "ml_xSetWindowAttributes_set_backing_planes"
external winAttr_set_backing_pixel: xSetWindowAttributes -> backing_pixel:uint -> unit = "ml_xSetWindowAttributes_set_backing_pixel"
external winAttr_set_save_under: xSetWindowAttributes -> save_under:bool -> unit = "ml_xSetWindowAttributes_set_save_under"
external winAttr_set_event_mask: xSetWindowAttributes -> event_mask:event_mask_list -> unit = "ml_xSetWindowAttributes_set_event_mask"
external winAttr_set_do_not_propagate_mask: xSetWindowAttributes -> do_not_propagate_mask:int -> unit = "ml_xSetWindowAttributes_set_do_not_propagate_mask"
external winAttr_set_override_redirect: xSetWindowAttributes -> override_redirect:bool -> unit = "ml_xSetWindowAttributes_set_override_redirect"
external winAttr_set_colormap: xSetWindowAttributes -> colormap:colormap -> unit = "ml_xSetWindowAttributes_set_colormap"
external winAttr_set_cursor: xSetWindowAttributes -> cursor:cursor -> unit = "ml_xSetWindowAttributes_set_cursor"

(* }}} *)


(** {3 Windows} *)

external xRootWindow: dpy:display -> screen_number -> window = "ml_XRootWindow"
external xDefaultRootWindow: dpy:display -> window = "ml_XDefaultRootWindow"

external xCreateSimpleWindow: dpy:display -> parent:window -> x:int -> y:int ->
                              width:uint -> height:uint -> border_width:uint ->
                              border:pixel_color -> background:pixel_color -> window
    = "ml_XCreateSimpleWindow_bytecode"
      "ml_XCreateSimpleWindow"

external xDestroyWindow: dpy:display -> win:window -> unit = "ml_XDestroyWindow"

external xStoreName: dpy:display -> win:window -> name:string -> unit = "ml_XStoreName"
external xFetchName: dpy:display -> win:window -> string = "ml_XFetchName" (* TODO test this function XXX *)

external xSelectInput: dpy:display -> win:window -> event_mask:event_mask list -> unit = "ml_XSelectInput"
external xMapWindow: dpy:display -> win:window -> unit = "ml_XMapWindow"
external xMapSubwindows: dpy:display -> win:window -> unit = "ml_XMapSubwindows"
external xMapRaised: dpy:display -> win:window -> unit = "ml_XMapRaised"


type window_class =
  | CopyFromParent
  | InputOutput
  | InputOnly

type winattr_valuemask =
  | CWBackPixmap
  | CWBackPixel
  | CWBorderPixmap
  | CWBorderPixel
  | CWBitGravity
  | CWWinGravity
  | CWBackingStore
  | CWBackingPlanes
  | CWBackingPixel
  | CWOverrideRedirect
  | CWSaveUnder
  | CWEventMask
  | CWDontPropagate
  | CWColormap
  | CWCursor

external xCreateWindow:
    dpy:display -> parent:window -> x:int -> y:int ->
    width:uint -> height:uint ->
    border_width:uint ->
    depth:int -> win_class:window_class -> visual:visual ->
    valuemask:winattr_valuemask list ->
    attributes:xSetWindowAttributes -> window
    = "ml_XCreateWindow_bytecode" "ml_XCreateWindow"


external xResizeWindow: dpy:display -> win:window -> width:uint -> height:uint -> unit = "ml_XResizeWindow"
external xMoveWindow: dpy:display -> win:window -> x:int -> y:int -> unit = "ml_XMoveWindow"
external xMoveResizeWindow: dpy:display -> win:window -> x:int -> y:int -> width:uint -> height:uint -> unit
    = "ml_XMoveResizeWindow_bytecode" "ml_XMoveResizeWindow"
external xLowerWindow: dpy:display -> win:window -> unit = "ml_XLowerWindow"
external xRaiseWindow: dpy:display -> win:window -> unit = "ml_XRaiseWindow"


(** {3 XScreen} *)

type xScreen
(** this type wraps the scruture Screen from the Xlib, but is here renamed with
    an additionnal x as a way of disambiguation with the type [screen_number] *)

external xDefaultScreenOfDisplay: dpy:display -> xScreen = "ml_XDefaultScreenOfDisplay"
(** the result points to a member of the input structure *)

external xScreenOfDisplay: dpy:display -> screen_number -> xScreen = "ml_XScreenOfDisplay"

external xDefaultVisualOfScreen: xScreen -> visual = "ml_XDefaultVisualOfScreen"
(** the result points to a member of the input structure *)

external xRootWindowOfScreen: xScreen -> window = "ml_XRootWindowOfScreen"
external xBlackPixelOfScreen: xScreen -> pixel_color = "ml_XBlackPixelOfScreen"
external xWhitePixelOfScreen: xScreen -> pixel_color = "ml_XWhitePixelOfScreen"
external xDefaultColormapOfScreen: xScreen -> colormap = "ml_XDefaultColormapOfScreen"
external xDefaultDepthOfScreen: xScreen -> int = "ml_XDefaultDepthOfScreen"
external xDefaultGCOfScreen: xScreen -> gc = "ml_XDefaultGCOfScreen"
external xDisplayOfScreen: xScreen -> display = "ml_XDisplayOfScreen"
external xWidthOfScreen: xScreen -> int = "ml_XWidthOfScreen"
external xHeightOfScreen: xScreen -> int = "ml_XHeightOfScreen"


(** {3 XSizeHints} *)

type xSizeHints
external new_xSizeHints: unit -> xSizeHints = "ml_alloc_XSizeHints"
(** this type is garbage collected *)

external xSizeHints_set_USPosition: xSizeHints -> x:int -> y:int -> unit = "ml_XSizeHints_set_USPosition"
external xSizeHints_set_USSize: xSizeHints -> width:int -> height:int -> unit = "ml_XSizeHints_set_USSize"
external xSizeHints_set_PPosition: xSizeHints -> x:int -> y:int -> unit = "ml_XSizeHints_set_PPosition"
external xSizeHints_set_PSize: xSizeHints -> width:int -> height:int -> unit = "ml_XSizeHints_set_PSize"
external xSizeHints_set_PMinSize: xSizeHints -> width:int -> height:int -> unit = "ml_XSizeHints_set_PMinSize"
external xSizeHints_set_PMaxSize: xSizeHints -> width:int -> height:int -> unit = "ml_XSizeHints_set_PMaxSize"
external xSizeHints_set_PResizeInc: xSizeHints -> width_inc:int -> height_inc:int -> unit = "ml_XSizeHints_set_PResizeInc"
external xSizeHints_set_PBaseSize: xSizeHints -> base_width:int -> base_height:int -> unit = "ml_XSizeHints_set_PBaseSize"
external xSizeHints_set_PAspect: xSizeHints -> min_aspect:int * int -> max_aspect:int * int -> unit = "ml_XSizeHints_set_PAspect"
external xSizeHints_set_PWinGravity: xSizeHints -> win_gravity:int -> unit = "ml_XSizeHints_set_PWinGravity"

external xSetNormalHints: dpy:display -> win:window -> hints:xSizeHints -> unit = "ml_XSetNormalHints"
external xSetStandardProperties: dpy:display -> win:window -> window_name:string -> icon_name:string ->
              icon_pixmap:pixmap option -> argv:string array -> hints:xSizeHints -> unit
    = "ml_XSetStandardProperties_bytecode"
      "ml_XSetStandardProperties"


(** {3 XEvents} *)

type 'a xEvent

type event_type =
  | KeyPress
  | KeyRelease
  | ButtonPress
  | ButtonRelease
  | MotionNotify
  | EnterNotify
  | LeaveNotify
  | FocusIn
  | FocusOut
  | KeymapNotify
  | Expose
  | GraphicsExpose
  | NoExpose
  | VisibilityNotify
  | CreateNotify
  | DestroyNotify
  | UnmapNotify
  | MapNotify
  | MapRequest
  | ReparentNotify
  | ConfigureNotify
  | ConfigureRequest
  | GravityNotify
  | ResizeRequest
  | CirculateNotify
  | CirculateRequest
  | PropertyNotify
  | SelectionClear
  | SelectionRequest
  | SelectionNotify
  | ColormapNotify
  | ClientMessage
  | MappingNotify

type any

type xKeyEvent
type xButtonEvent
type xMotionEvent
type xCrossingEvent
type xFocusChangeEvent
type xExposeEvent
type xGraphicsExposeEvent
type xNoExposeEvent
type xVisibilityEvent
type xCreateWindowEvent
type xDestroyWindowEvent
type xUnmapEvent
type xMapEvent
type xMapRequestEvent
type xReparentEvent
type xConfigureEvent
type xGravityEvent
type xResizeRequestEvent
type xConfigureRequestEvent
type xCirculateEvent
type xCirculateRequestEvent
type xPropertyEvent
type xSelectionClearEvent
type xSelectionRequestEvent
type xSelectionEvent
type xColormapEvent
type xClientMessageEvent
type xMappingEvent
type xErrorEvent
type xKeymapEvent

type xKeyPressedEvent  = xKeyEvent
type xKeyReleasedEvent = xKeyEvent

type xButtonPressedEvent  = xButtonEvent
type xButtonReleasedEvent = xButtonEvent

type xEnterWindowEvent = xCrossingEvent
type xLeaveWindowEvent = xCrossingEvent

type xFocusInEvent  = xFocusChangeEvent
type xFocusOutEvent = xFocusChangeEvent


type event_kind =
  | XKeyPressedEvent       of xKeyPressedEvent       xEvent
  | XKeyReleasedEvent      of xKeyReleasedEvent      xEvent

  | XButtonPressedEvent    of xButtonPressedEvent    xEvent
  | XButtonReleasedEvent   of xButtonReleasedEvent   xEvent

  | XMotionEvent           of xMotionEvent           xEvent
  | XCrossingEvent         of xCrossingEvent         xEvent
  | XFocusChangeEvent      of xFocusChangeEvent      xEvent
  | XExposeEvent           of xExposeEvent           xEvent
  | XGraphicsExposeEvent   of xGraphicsExposeEvent   xEvent
  | XNoExposeEvent         of xNoExposeEvent         xEvent
  | XVisibilityEvent       of xVisibilityEvent       xEvent
  | XCreateWindowEvent     of xCreateWindowEvent     xEvent
  | XDestroyWindowEvent    of xDestroyWindowEvent    xEvent
  | XUnmapEvent            of xUnmapEvent            xEvent
  | XMapEvent              of xMapEvent              xEvent
  | XMapRequestEvent       of xMapRequestEvent       xEvent
  | XReparentEvent         of xReparentEvent         xEvent
  | XConfigureEvent        of xConfigureEvent        xEvent
  | XGravityEvent          of xGravityEvent          xEvent
  | XResizeRequestEvent    of xResizeRequestEvent    xEvent
  | XConfigureRequestEvent of xConfigureRequestEvent xEvent
  | XCirculateEvent        of xCirculateEvent        xEvent
  | XCirculateRequestEvent of xCirculateRequestEvent xEvent
  | XPropertyEvent         of xPropertyEvent         xEvent
  | XSelectionClearEvent   of xSelectionClearEvent   xEvent
  | XSelectionRequestEvent of xSelectionRequestEvent xEvent
  | XSelectionEvent        of xSelectionEvent        xEvent
  | XColormapEvent         of xColormapEvent         xEvent
  | XClientMessageEvent    of xClientMessageEvent    xEvent
  | XMappingEvent          of xMappingEvent          xEvent
  | XErrorEvent            of xErrorEvent            xEvent
  | XKeymapEvent           of xKeymapEvent           xEvent


external new_xEvent: unit -> any xEvent = "ml_alloc_XEvent"
(** the return value is garbage collected *)

external xNextEvent: dpy:display -> event:any xEvent -> unit = "ml_XNextEvent"
external xNextEventFun: dpy:display -> any xEvent = "ml_XNextEvent_fun"
external xWindowEvent: dpy:display -> win:window -> event_mask list -> 'a xEvent = "ml_XWindowEvent"
external xPending: dpy:display -> int = "ml_XPending"

type queued_mode =
  | QueuedAlready
  | QueuedAfterFlush
  | QueuedAfterReading

external xEventsQueued: dpy:display -> mode:queued_mode -> int = "ml_XEventsQueued"

external xCheckTypedEvent: dpy:display -> event_type -> 'a xEvent -> bool = "ml_XCheckTypedEvent"

external xEventType: event:'a xEvent -> event_type = "ml_XEvent_type"


#if defined(MLI)
val xEventKind : event:'a xEvent -> event_kind
(** selects the right type of the event *)
#else
let xEventKind ~event =
  match xEventType ~event with
  | MotionNotify        -> XMotionEvent           (Obj.magic event : xMotionEvent           xEvent)
 
  | KeyPress            -> XKeyPressedEvent       (Obj.magic event : xKeyPressedEvent       xEvent)
  | KeyRelease          -> XKeyReleasedEvent      (Obj.magic event : xKeyReleasedEvent      xEvent)

  | ButtonPress         -> XButtonPressedEvent    (Obj.magic event : xButtonPressedEvent    xEvent)
  | ButtonRelease       -> XButtonReleasedEvent   (Obj.magic event : xButtonReleasedEvent   xEvent)
 
  (*
  | EnterNotify
  | LeaveNotify         -> XCrossingEvent         (Obj.magic event : xCrossingEvent         xEvent)
  *)
  | EnterNotify         -> XCrossingEvent         (Obj.magic event : xEnterWindowEvent      xEvent)
  | LeaveNotify         -> XCrossingEvent         (Obj.magic event : xLeaveWindowEvent      xEvent)
 
  (*
  | FocusIn
  | FocusOut            -> XFocusChangeEvent      (Obj.magic event : xFocusChangeEvent      xEvent)
  *)
  | FocusIn             -> XFocusChangeEvent      (Obj.magic event : xFocusInEvent          xEvent)
  | FocusOut            -> XFocusChangeEvent      (Obj.magic event : xFocusOutEvent         xEvent)
 
  | KeymapNotify        -> XKeymapEvent           (Obj.magic event : xKeymapEvent           xEvent)
  | Expose              -> XExposeEvent           (Obj.magic event : xExposeEvent           xEvent)
  | GraphicsExpose      -> XGraphicsExposeEvent   (Obj.magic event : xGraphicsExposeEvent   xEvent)
  | NoExpose            -> XNoExposeEvent         (Obj.magic event : xNoExposeEvent         xEvent)
  | VisibilityNotify    -> XVisibilityEvent       (Obj.magic event : xVisibilityEvent       xEvent)
  | CreateNotify        -> XCreateWindowEvent     (Obj.magic event : xCreateWindowEvent     xEvent)
  | DestroyNotify       -> XDestroyWindowEvent    (Obj.magic event : xDestroyWindowEvent    xEvent)
  | UnmapNotify         -> XUnmapEvent            (Obj.magic event : xUnmapEvent            xEvent)
  | MapNotify           -> XMapEvent              (Obj.magic event : xMapEvent              xEvent)
  | MapRequest          -> XMapRequestEvent       (Obj.magic event : xMapRequestEvent       xEvent)
  | ReparentNotify      -> XReparentEvent         (Obj.magic event : xReparentEvent         xEvent)
  | ConfigureNotify     -> XConfigureEvent        (Obj.magic event : xConfigureEvent        xEvent)
  | ConfigureRequest    -> XConfigureRequestEvent (Obj.magic event : xConfigureRequestEvent xEvent)
  | GravityNotify       -> XGravityEvent          (Obj.magic event : xGravityEvent          xEvent)
  | ResizeRequest       -> XResizeRequestEvent    (Obj.magic event : xResizeRequestEvent    xEvent)
  | CirculateNotify     -> XCirculateEvent        (Obj.magic event : xCirculateEvent        xEvent)
  | CirculateRequest    -> XCirculateRequestEvent (Obj.magic event : xCirculateRequestEvent xEvent)
  | PropertyNotify      -> XPropertyEvent         (Obj.magic event : xPropertyEvent         xEvent)
  | SelectionClear      -> XSelectionClearEvent   (Obj.magic event : xSelectionClearEvent   xEvent)
  | SelectionRequest    -> XSelectionRequestEvent (Obj.magic event : xSelectionRequestEvent xEvent)
  | SelectionNotify     -> XSelectionEvent        (Obj.magic event : xSelectionEvent        xEvent)
  | ColormapNotify      -> XColormapEvent         (Obj.magic event : xColormapEvent         xEvent)
  | ClientMessage       -> XClientMessageEvent    (Obj.magic event : xClientMessageEvent    xEvent)
  | MappingNotify       -> XMappingEvent          (Obj.magic event : xMappingEvent          xEvent)
;;
#endif

(* {{{ XEvent conversion *)
(** {4 XEvent Conversions} *)

#if defined(ML)

let to_xMotionEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> MotionNotify then
    invalid_arg "to_xMotionEvent";
  (Obj.magic event : xMotionEvent  xEvent)
;;

let to_xKeyEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> KeyPress &&
     (xEventType ~event) <> KeyRelease then
    invalid_arg "to_xKeyEvent";
  (Obj.magic event : xKeyEvent  xEvent)
;;

let to_xKeyPressedEvent ~(event : 'a xEvent) =
  let t = (xEventType ~event) in
  if (xEventType ~event) <> KeyPress then
    begin
      Printf.eprintf "Event Type: %d\n%!" (Obj.magic t : int);
      if t = Expose then
        Printf.printf "10 confirmed\n%!";
      invalid_arg "to_xKeyPressedEvent";
    end;
  (Obj.magic event : xKeyPressedEvent  xEvent)
;;

let to_xKeyReleasedEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> KeyRelease then
    invalid_arg "to_xKeyReleasedEvent";
  (Obj.magic event : xKeyReleasedEvent  xEvent)
;;

let to_xButtonEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> ButtonPress &&
     (xEventType ~event) <> ButtonRelease then
    invalid_arg "to_xButtonEvent";
  (Obj.magic event : xButtonEvent  xEvent)
;;

let to_xButtonPressedEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> ButtonPress then
    invalid_arg "to_xButtonPressedEvent";
  (Obj.magic event : xButtonPressedEvent  xEvent)
;;

let to_xButtonReleasedEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> ButtonRelease then
    invalid_arg "to_xButtonReleasedEvent";
  (Obj.magic event : xButtonReleasedEvent  xEvent)
;;

let to_xCrossingEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> EnterNotify &&
     (xEventType ~event) <> LeaveNotify then
    invalid_arg "to_xCrossingEvent";
  (Obj.magic event : xCrossingEvent  xEvent)
;;

let to_xEnterWindowEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> EnterNotify then
    invalid_arg "to_xEnterWindowEvent";
  (Obj.magic event : xEnterWindowEvent  xEvent)
;;

let to_xLeaveWindowEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> LeaveNotify then
    invalid_arg "to_xLeaveWindowEvent";
  (Obj.magic event : xLeaveWindowEvent  xEvent)
;;

let to_xFocusChangeEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> FocusIn &&
     (xEventType ~event) <> FocusOut then
    invalid_arg "to_xFocusChangeEvent";
  (Obj.magic event : xFocusChangeEvent  xEvent)
;;

let to_xFocusInEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> FocusIn then
    invalid_arg "to_xFocusInEvent";
  (Obj.magic event : xFocusInEvent  xEvent)
;;

let to_xFocusOutEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> FocusOut then
    invalid_arg "to_xFocusOutEvent";
  (Obj.magic event : xFocusOutEvent  xEvent)
;;

let to_xKeymapEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> KeymapNotify then
    invalid_arg "to_xKeymapEvent";
  (Obj.magic event : xKeymapEvent  xEvent)
;;

let to_xExposeEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> Expose then
    invalid_arg "to_xExposeEvent";
  (Obj.magic event : xExposeEvent  xEvent)
;;

let to_xGraphicsExposeEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> GraphicsExpose then
    invalid_arg "to_xGraphicsExposeEvent";
  (Obj.magic event : xGraphicsExposeEvent  xEvent)
;;

let to_xNoExposeEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> NoExpose then
    invalid_arg "to_xNoExposeEvent";
  (Obj.magic event : xNoExposeEvent  xEvent)
;;

let to_xVisibilityEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> VisibilityNotify then
    invalid_arg "to_xVisibilityEvent";
  (Obj.magic event : xVisibilityEvent  xEvent)
;;

let to_xCreateWindowEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> CreateNotify then
    invalid_arg "to_xCreateWindowEvent";
  (Obj.magic event : xCreateWindowEvent  xEvent)
;;

let to_xDestroyWindowEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> DestroyNotify then
    invalid_arg "to_xDestroyWindowEvent";
  (Obj.magic event : xDestroyWindowEvent  xEvent)
;;

let to_xUnmapEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> UnmapNotify then
    invalid_arg "to_xUnmapEvent";
  (Obj.magic event : xUnmapEvent  xEvent)
;;

let to_xMapEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> MapNotify then
    invalid_arg "to_xMapEvent";
  (Obj.magic event : xMapEvent  xEvent)
;;

let to_xMapRequestEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> MapRequest then
    invalid_arg "to_xMapRequestEvent";
  (Obj.magic event : xMapRequestEvent  xEvent)
;;

let to_xReparentEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> ReparentNotify then
    invalid_arg "to_xReparentEvent";
  (Obj.magic event : xReparentEvent  xEvent)
;;

let to_xConfigureEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> ConfigureNotify then
    invalid_arg "to_xConfigureEvent";
  (Obj.magic event : xConfigureEvent  xEvent)
;;

let to_xConfigureRequestEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> ConfigureRequest then
    invalid_arg "to_xConfigureRequestEvent";
  (Obj.magic event : xConfigureRequestEvent  xEvent)
;;

let to_xGravityEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> GravityNotify then
    invalid_arg "to_xGravityEvent";
  (Obj.magic event : xGravityEvent  xEvent)
;;

let to_xResizeRequestEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> ResizeRequest then
    invalid_arg "to_xResizeRequestEvent";
  (Obj.magic event : xResizeRequestEvent  xEvent)
;;

let to_xCirculateEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> CirculateNotify then
    invalid_arg "to_xCirculateEvent";
  (Obj.magic event : xCirculateEvent  xEvent)
;;

let to_xCirculateRequestEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> CirculateRequest then
    invalid_arg "to_xCirculateRequestEvent";
  (Obj.magic event : xCirculateRequestEvent  xEvent)
;;

let to_xPropertyEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> PropertyNotify then
    invalid_arg "to_xPropertyEvent";
  (Obj.magic event : xPropertyEvent  xEvent)
;;

let to_xSelectionClearEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> SelectionClear then
    invalid_arg "to_xSelectionClearEvent";
  (Obj.magic event : xSelectionClearEvent  xEvent)
;;

let to_xSelectionRequestEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> SelectionRequest then
    invalid_arg "to_xSelectionRequestEvent";
  (Obj.magic event : xSelectionRequestEvent  xEvent)
;;

let to_xSelectionEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> SelectionNotify then
    invalid_arg "to_xSelectionEvent";
  (Obj.magic event : xSelectionEvent  xEvent)
;;

let to_xColormapEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> ColormapNotify then
    invalid_arg "to_xColormapEvent";
  (Obj.magic event : xColormapEvent  xEvent)
;;

let to_xClientMessageEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> ClientMessage then
    invalid_arg "to_xClientMessageEvent";
  (Obj.magic event : xClientMessageEvent  xEvent)
;;

let to_xMappingEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> MappingNotify then
    invalid_arg "to_xMappingEvent";
  (Obj.magic event : xMappingEvent  xEvent)
;;

#else

(** convertions from type [any xEvent] *)

val to_xMotionEvent : event:'a xEvent -> xMotionEvent xEvent
val to_xKeyEvent : event:'a xEvent -> xKeyEvent xEvent
val to_xKeyPressedEvent : event:'a xEvent -> xKeyPressedEvent xEvent
val to_xKeyReleasedEvent : event:'a xEvent -> xKeyReleasedEvent xEvent
val to_xButtonEvent : event:'a xEvent -> xButtonEvent xEvent
val to_xButtonPressedEvent : event:'a xEvent -> xButtonPressedEvent xEvent
val to_xButtonReleasedEvent : event:'a xEvent -> xButtonReleasedEvent xEvent
val to_xCrossingEvent : event:'a xEvent -> xCrossingEvent xEvent
val to_xEnterWindowEvent : event:'a xEvent -> xEnterWindowEvent xEvent
val to_xLeaveWindowEvent : event:'a xEvent -> xLeaveWindowEvent xEvent
val to_xFocusChangeEvent : event:'a xEvent -> xFocusChangeEvent xEvent
val to_xFocusInEvent : event:'a xEvent -> xFocusInEvent xEvent
val to_xFocusOutEvent : event:'a xEvent -> xFocusOutEvent xEvent
val to_xKeymapEvent : event:'a xEvent -> xKeymapEvent xEvent
val to_xExposeEvent : event:'a xEvent -> xExposeEvent xEvent
val to_xGraphicsExposeEvent : event:'a xEvent -> xGraphicsExposeEvent xEvent
val to_xNoExposeEvent : event:'a xEvent -> xNoExposeEvent xEvent
val to_xVisibilityEvent : event:'a xEvent -> xVisibilityEvent xEvent
val to_xCreateWindowEvent : event:'a xEvent -> xCreateWindowEvent xEvent
val to_xDestroyWindowEvent : event:'a xEvent -> xDestroyWindowEvent xEvent
val to_xUnmapEvent : event:'a xEvent -> xUnmapEvent xEvent
val to_xMapEvent : event:'a xEvent -> xMapEvent xEvent
val to_xMapRequestEvent : event:'a xEvent -> xMapRequestEvent xEvent
val to_xReparentEvent : event:'a xEvent -> xReparentEvent xEvent
val to_xConfigureEvent : event:'a xEvent -> xConfigureEvent xEvent
val to_xConfigureRequestEvent : event:'a xEvent -> xConfigureRequestEvent xEvent
val to_xGravityEvent : event:'a xEvent -> xGravityEvent xEvent
val to_xResizeRequestEvent : event:'a xEvent -> xResizeRequestEvent xEvent
val to_xCirculateEvent : event:'a xEvent -> xCirculateEvent xEvent
val to_xCirculateRequestEvent : event:'a xEvent -> xCirculateRequestEvent xEvent
val to_xPropertyEvent : event:'a xEvent -> xPropertyEvent xEvent
val to_xSelectionClearEvent : event:'a xEvent -> xSelectionClearEvent xEvent
val to_xSelectionRequestEvent : event:'a xEvent -> xSelectionRequestEvent xEvent
val to_xSelectionEvent : event:'a xEvent -> xSelectionEvent xEvent
val to_xColormapEvent : event:'a xEvent -> xColormapEvent xEvent
val to_xClientMessageEvent : event:'a xEvent -> xClientMessageEvent xEvent
val to_xMappingEvent : event:'a xEvent -> xMappingEvent xEvent
#endif

(* }}} *)
(* {{{ XEvents contents *)
(** {4 XEvents Contents} *)

type xAnyEvent_contents = {
    xany_type: event_type;
    xany_serial: uint;
    xany_send_event: bool;
    xany_display: display;
    xany_window: window;
  }
external xAnyEvent_datas: 'a xEvent -> xAnyEvent_contents = "ml_XAnyEvent_datas"

type time = int64

type logical_state =
  | AnyModifier
  | Button1Mask
  | Button2Mask
  | Button3Mask
  | Button4Mask
  | Button5Mask
  | ShiftMask
  | LockMask
  | ControlMask
  | Mod1Mask
  | Mod2Mask
  | Mod3Mask
  | Mod4Mask
  | Mod5Mask

type xKeyEvent_contents = {
    key_time: time;
    key_x: int;
    key_y: int;
    key_x_root: int;
    key_y_root: int;
    key_state: logical_state;
    key_keycode: uint;
    key_same_screen: bool;
  }
external xKeyEvent_datas: xKeyEvent xEvent -> xKeyEvent_contents = "ml_XKeyEvent_datas"

type xConfigureEvent_contents = {
    conf_x: int;
    conf_y: int;
    conf_width: int;
    conf_height: int;
    conf_border_width: int;
    conf_above: window;
    conf_override_redirect: bool;
  }
external xConfigureEvent_datas: xConfigureEvent xEvent -> xConfigureEvent_contents = "ml_XConfigureEvent_datas"

(*
type button_mask =
  | Button1Mask
  | Button2Mask
  | Button3Mask
  | Button4Mask
  | Button5Mask
*)

type xMotionEvent_contents = {
    motion_root: window;
    motion_subwindow: window;
    motion_time: time;
    motion_x: int;
    motion_y: int;
    motion_x_root: int;
    motion_y_root: int;
    motion_state: (* button_mask *) logical_state list; (* TODO check the additionnal Masks *)
    motion_is_hint: char;
    motion_same_screen: bool;
  }
external xMotionEvent_datas: xMotionEvent xEvent -> xMotionEvent_contents = "ml_XMotionEvent_datas"

type button =
  | AnyButton
  | Button1
  | Button2
  | Button3
  | Button4
  | Button5

type xButtonEvent_contents = {
    button_root: window;
    button_subwindow: window;
    button_time: time;
    button_x: int;
    button_y: int;
    button_x_root: int;
    button_y_root: int;
    button_state: uint;
    button: button;
    button_same_screen: bool;
  }
external xButtonEvent_datas: xButtonEvent xEvent -> xButtonEvent_contents = "ml_XButtonEvent_datas"

(* }}} *)

(* external _xLookupString: event:xKeyEvent xEvent -> unit = "_ml_XLookupString" *)

type keysym = int
external xLookupString: event:xKeyEvent xEvent -> buffer:string -> int * keysym = "ml_XLookupString"

external xLookupKeysym: event:xKeyEvent xEvent -> index:int -> keysym = "ml_XLookupKeysym"
external xKeycodeToKeysym: dpy:display -> keycode:uint -> index:int -> keysym = "ml_XKeycodeToKeysym"

external xRefreshKeyboardMapping : event:xMappingEvent xEvent -> unit = "ml_XRefreshKeyboardMapping"


(** {3 Atoms} *)

type atom
external xSetWMProtocols: dpy:display -> win:window -> protocols:atom -> count:int -> unit = "ml_XSetWMProtocols"
external xInternAtom: dpy:display -> atom_name:string -> only_if_exists:bool -> atom = "ml_XInternAtom"
(** if None is returned, raises [Not_found] *)

external xGetAtomName: dpy:display -> atom:atom -> string = "ml_XGetAtomName"

external xEvent_xclient_data : xClientMessageEvent xEvent -> atom = "ml_XEvent_xclient_data_l_0"



(** {3 Font} *)
(* xlib.pdf page 155 *)

type font
type xFontStruct  (* pointer to a structure *)

external xLoadFont: dpy:display -> name:string -> font = "ml_XLoadFont"
external xLoadQueryFont: dpy:display -> name:string -> xFontStruct = "ml_XLoadQueryFont"
external xQueryFont: dpy:display -> font:font -> xFontStruct = "ml_XQueryFont"
external xQueryFontGC: dpy:display -> gc:gc -> xFontStruct = "ml_XQueryFontGC"
external xSetFont: dpy:display -> gc:gc -> font:font -> unit = "ml_XSetFont"

external xFontStruct_font: xFontStruct -> font = "ml_XFontStruct_get_fid"
external xFontStruct_ascent: xFontStruct -> int = "ml_XFontStruct_get_ascent"
external xFontStruct_descent: xFontStruct -> int = "ml_XFontStruct_get_descent"
external xFontStruct_all_chars_exist: xFontStruct -> bool = "ml_XFontStruct_get_all_chars_exist"

external xFontStruct_get_height: xFontStruct -> int * int = "ml_XFontStruct_get_height"
(** returns the [(ascent, descent)] pair *)

external xTextWidth: xFontStruct -> string -> int = "ml_XTextWidth"

type xCharStruct = {
    lbearing: int;
    rbearing: int;
    width: int;
    ascent: int;
    descent: int;
  }

external xFontStruct_min_bounds: xFontStruct -> xCharStruct = "ml_xFontStruct_min_bounds"
external xFontStruct_max_bounds: xFontStruct -> xCharStruct = "ml_xFontStruct_max_bounds"


(** {3 Graphics Context} *)

external xDefaultGC: dpy:display -> screen_number -> gc = "ml_XDefaultGC"

type xGCValues
external new_xGCValues: unit -> xGCValues = "ml_alloc_XGCValues"
(** the return value is garbage collected *)

type line_style =
  | LineSolid
  | LineOnOffDash
  | LineDoubleDash

type cap_style =
  | CapNotLast
  | CapButt
  | CapRound
  | CapProjecting

type join_style =
  | JoinMiter
  | JoinRound
  | JoinBevel

type fill_style =
  | FillSolid
  | FillTiled
  | FillStippled
  | FillOpaqueStippled

type fill_rule =
  | EvenOddRule
  | WindingRule

type logop_func =
  | GXclear
  | GXand
  | GXandReverse
  | GXcopy
  | GXandInverted
  | GXnoop
  | GXxor
  | GXor
  | GXnor
  | GXequiv
  | GXinvert
  | GXorReverse
  | GXcopyInverted
  | GXorInverted
  | GXnand
  | GXset

type arc_mode =
  | ArcChord
  | ArcPieSlice

type subwindow_mode =
  | ClipByChildren
  | IncludeInferiors

external xGCValues_set_foreground: gcv:xGCValues -> fg:pixel_color -> unit = "ml_XGCValues_set_foreground"
external xGCValues_set_background: gcv:xGCValues -> bg:pixel_color -> unit = "ml_XGCValues_set_background"
external xGCValues_set_graphics_exposures: gcv:xGCValues -> bool -> unit = "ml_XGCValues_set_graphics_exposures"
external xGCValues_set_tile: gcv:xGCValues -> pixmap -> unit = "ml_XGCValues_set_tile"
external xGCValues_set_clip_x_origin: gcv:xGCValues -> int -> unit = "ml_XGCValues_set_clip_x_origin"
external xGCValues_set_clip_y_origin: gcv:xGCValues -> int -> unit = "ml_XGCValues_set_clip_y_origin"
external xGCValues_set_ts_x_origin: gcv:xGCValues -> int -> unit = "ml_XGCValues_set_ts_x_origin"
external xGCValues_set_ts_y_origin: gcv:xGCValues -> int -> unit = "ml_XGCValues_set_ts_y_origin"
external xGCValues_set_line_style: gcv:xGCValues -> line_style -> unit = "ml_XGCValues_set_line_style"
external xGCValues_set_cap_style: gcv:xGCValues -> cap_style -> unit = "ml_XGCValues_set_cap_style"
external xGCValues_set_join_style: gcv:xGCValues -> join_style -> unit = "ml_XGCValues_set_join_style"
external xGCValues_set_fill_style: gcv:xGCValues -> fill_style -> unit = "ml_XGCValues_set_fill_style"
external xGCValues_set_fill_rule: gcv:xGCValues -> fill_rule -> unit = "ml_XGCValues_set_fill_rule"
external xGCValues_set_function: gcv:xGCValues -> logop_func -> unit = "ml_XGCValues_set_function"
external xGCValues_set_line_width: gcv:xGCValues -> int -> unit = "ml_XGCValues_set_line_width"
external xGCValues_set_arc_mode: gcv:xGCValues -> arc_mode -> unit = "ml_XGCValues_set_arc_mode"
external xGCValues_set_font: gcv:xGCValues -> font -> unit = "ml_XGCValues_set_font"
external xGCValues_set_subwindow_mode: gcv:xGCValues -> subwindow_mode -> unit = "ml_XGCValues_set_subwindow_mode"


external xGCValues_get_foreground: gcv:xGCValues -> pixel_color = "ml_XGCValues_get_foreground"
external xGCValues_get_background: gcv:xGCValues -> pixel_color = "ml_XGCValues_get_background"
external xGCValues_get_graphics_exposures: gcv:xGCValues -> bool = "ml_XGCValues_get_graphics_exposures"
external xGCValues_get_tile: gcv:xGCValues -> pixmap = "ml_XGCValues_get_tile"
external xGCValues_get_clip_x_origin: gcv:xGCValues -> int = "ml_XGCValues_get_clip_x_origin"
external xGCValues_get_clip_y_origin: gcv:xGCValues -> int = "ml_XGCValues_get_clip_y_origin"
external xGCValues_get_ts_x_origin: gcv:xGCValues -> int = "ml_XGCValues_get_ts_x_origin"
external xGCValues_get_ts_y_origin: gcv:xGCValues -> int = "ml_XGCValues_get_ts_y_origin"
external xGCValues_get_line_style: gcv:xGCValues -> line_style = "ml_XGCValues_get_line_style"
external xGCValues_get_cap_style: gcv:xGCValues -> cap_style = "ml_XGCValues_get_cap_style"
external xGCValues_get_join_style: gcv:xGCValues -> join_style = "ml_XGCValues_get_join_style"
external xGCValues_get_fill_style: gcv:xGCValues -> fill_style = "ml_XGCValues_get_fill_style"
external xGCValues_get_fill_rule: gcv:xGCValues -> fill_rule = "ml_XGCValues_get_fill_rule"
external xGCValues_get_function: gcv:xGCValues -> logop_func = "ml_XGCValues_get_function"
external xGCValues_get_line_width: gcv:xGCValues -> int = "ml_XGCValues_get_line_width"
external xGCValues_get_arc_mode: gcv:xGCValues -> arc_mode = "ml_XGCValues_get_arc_mode"
external xGCValues_get_font: gcv:xGCValues -> font = "ml_XGCValues_get_font"
external xGCValues_get_subwindow_mode: gcv:xGCValues -> subwindow_mode = "ml_XGCValues_get_subwindow_mode"


type gc_valuemask =
  | GCFunction
  | GCPlaneMask
  | GCForeground
  | GCBackground
  | GCLineWidth
  | GCLineStyle
  | GCCapStyle
  | GCJoinStyle
  | GCFillStyle
  | GCFillRule
  | GCTile
  | GCStipple
  | GCTileStipXOrigin
  | GCTileStipYOrigin
  | GCFont
  | GCSubwindowMode
  | GCGraphicsExposures
  | GCClipXOrigin
  | GCClipYOrigin
  | GCClipMask
  | GCDashOffset
  | GCDashList
  | GCArcMode


#if defined(ML)
external xCreateGC: dpy:display -> d:'a drawable -> gc_valuemask list -> xGCValues -> gc = "ml_XCreateGC"

external do_finalize_gc: gc -> unit = "do_finalize_GC"
let xCreateGC ~dpy ~d mask vals =
  let gc = xCreateGC ~dpy ~d mask vals in
  Gc.finalise do_finalize_gc gc;
  (gc)
;;
#else
val xCreateGC: dpy:display -> d:'a drawable -> gc_valuemask list -> xGCValues -> gc
#endif


external xChangeGC: dpy:display -> gc:gc -> gc_valuemask list -> xGCValues -> unit = "ml_XChangeGC"
external xGetGCValues: dpy:display -> gc:gc -> gc_valuemask list -> xGCValues = "ml_XGetGCValues"

external xSetLineAttributes: dpy:display -> gc:gc -> line_width:uint -> line_style:line_style ->
                             cap_style:cap_style -> join_style:join_style -> unit
    = "ml_XSetLineAttributes_bytecode"
      "ml_XSetLineAttributes"

external xSetFillStyle: dpy:display -> gc:gc -> fill_style:fill_style -> unit = "ml_XSetFillStyle"



(** {3 Drawing} *)

external xSetForeground: dpy:display -> gc:gc -> fg:pixel_color -> unit = "ml_XSetForeground"
external xSetBackground: dpy:display -> gc:gc -> bg:pixel_color -> unit = "ml_XSetBackground"

external xClearWindow: dpy:display -> win:window -> unit = "ml_XClearWindow"
external xClearArea: dpy:display -> win:window -> x:int -> y:int ->
                     width:uint -> height:uint -> exposures:bool -> unit
    = "ml_XClearArea_bytecode"
      "ml_XClearArea"


external xDrawPoint: dpy:display -> d:'a drawable -> gc:gc -> x:int -> y:int -> unit = "ml_XDrawPoint"

type coordinate_mode =
  | CoordModeOrigin
  | CoordModePrevious

type x_point = { pnt_x: int; pnt_y: int }

external xDrawPoints: dpy:display -> d:'a drawable -> gc:gc -> points: x_point array -> mode:coordinate_mode -> unit
    = "ml_XDrawPoints"

external xDrawLine: dpy:display -> d:'a drawable -> gc:gc -> x1:int -> y1:int -> x2:int -> y2:int -> unit
    = "ml_XDrawLine_bytecode"
      "ml_XDrawLine"

external xDrawLines: dpy:display -> d:'a drawable -> gc:gc -> points: x_point array -> mode:coordinate_mode -> unit
    = "ml_XDrawLines"

type segment = {
    x1:int;
    y1:int;
    x2:int;
    y2:int;
  }
external xDrawSegments: dpy:display -> d:'a drawable -> gc:gc -> segments:segment array -> unit = "ml_XDrawSegments"

external xDrawRectangle: dpy:display -> d:'a drawable -> gc:gc ->
                         x:int -> y:int -> width:uint -> height:uint -> unit
    = "ml_XDrawRectangle_bytecode"
      "ml_XDrawRectangle"

external xFillRectangle: dpy:display -> d:'a drawable -> gc:gc ->
                         x:int -> y:int -> width:uint -> height:uint -> unit
    = "ml_XFillRectangle_bytecode"
      "ml_XFillRectangle"

type x_rectangle = {
    rect_x : int;
    rect_y : int;
    rect_width : uint;
    rect_height : uint;
  }
external xFillRectangles: dpy:display -> d:'a drawable -> gc:gc -> rectangles: x_rectangle array -> unit
    = "ml_XFillRectangles"

external xDrawRectangles: dpy:display -> d:'a drawable -> gc:gc -> rectangles: x_rectangle array -> unit
    = "ml_XDrawRectangles"

external xDrawArc: dpy:display -> d:'a drawable -> gc:gc -> x:int -> y:int ->
                   width:uint -> height:uint -> angle1:int -> angle2:int -> unit
    = "ml_XDrawArc_bytecode"
      "ml_XDrawArc"

external xFillArc: dpy:display -> d:'a drawable -> gc:gc -> x:int -> y:int ->
                   width:uint -> height:uint -> angle1:int -> angle2:int -> unit
    = "ml_XFillArc_bytecode"
      "ml_XFillArc"

type x_arc = {
    arc_x : int;
    arc_y : int;
    arc_width : uint;
    arc_height : uint;
    arc_angle1 : int;
    arc_angle2 : int;
  }
external xDrawArcs: dpy:display -> d:'a drawable -> gc:gc -> arcs: x_arc array -> unit = "ml_XDrawArcs"
external xFillArcs: dpy:display -> d:'a drawable -> gc:gc -> arcs: x_arc array -> unit = "ml_XFillArcs"

type shape_kind =
  | Complex
  | Nonconvex
  | Convex

external xFillPolygon: dpy:display -> d:'a drawable -> gc:gc -> points: x_point array -> 
                       shape:shape_kind -> mode:coordinate_mode -> unit
    = "ml_XFillPolygon_bytecode"
      "ml_XFillPolygon"

external xDrawString: dpy:display -> d:'a drawable -> gc:gc ->
                      x:int -> y:int -> str:string -> unit
    = "ml_XDrawString_bytecode"
      "ml_XDrawString"

external xDrawImageString: dpy:display -> d:'a drawable -> gc:gc ->
                      x:int -> y:int -> str:string -> unit
    = "ml_XDrawImageString_bytecode"
      "ml_XDrawImageString"


(** {3 Drawable} *)

external xCopyArea: dpy:display ->
    src:'a drawable ->
    dest:'b drawable ->
    gc:gc ->
    src_x:int ->
    src_y:int ->
    width:uint ->
    height: uint ->
    dest_x:int ->
    dest_y:int -> unit
    = "ml_XCopyArea_bytecode"
      "ml_XCopyArea"

external xCreatePixmap: dpy:display -> d:'a drawable -> width:uint -> height:uint -> depth:uint -> pixmap
    = "ml_XCreatePixmap"
external xCreateBitmapFromData: dpy:display -> d:'a drawable -> data:string -> width:uint -> height:uint -> pixmap
    = "ml_XCreateBitmapFromData"

external xCreatePixmapCursor: dpy:display -> source:pixmap -> mask:pixmap ->
    foreground:xColor -> background:xColor -> x:uint -> y:uint -> cursor
    = "ml_XCreatePixmapCursor_bytecode"
      "ml_XCreatePixmapCursor"

external xFreePixmap: dpy:display -> pixmap -> unit = "ml_XFreePixmap"

external xQueryBestTile: dpy:display -> d:'a drawable -> width:uint -> height:uint -> uint * uint = "ml_XQueryBestTile"

type xPixmapFormatValues = {
    pxm_depth: int;
    bits_per_pixel: int;
    scanline_pad: int;
  }

external xListPixmapFormats: dpy:display -> xPixmapFormatValues array = "ml_XListPixmapFormats"
external xBitmapUnit: dpy:display -> int = "ml_XBitmapUnit"
external xBitmapPad: dpy:display -> int = "ml_XBitmapPad"


(** {4 XImage} *)

type byte_order = LSBFirst | MSBFirst
external xImageByteOrder: dpy:display -> byte_order = "ml_XImageByteOrder"

type ximage_format = XYBitmap | XYPixmap | ZPixmap

type xImage
external xCreateImage:
    dpy:display -> visual:visual -> depth:int -> fmt:ximage_format -> offset:int ->
    data:'a -> width:uint -> height:uint -> bitmap_pad:int -> bytes_per_line:int -> xImage
    = "ml_XCreateImage_bytecode"
      "ml_XCreateImage"
(** [data] can be a string or a bigarray (ala glcaml) *)

external xDestroyImage: ximage:xImage -> unit = "ml_XDestroyImage"


(** {4 Keyboard} *)

external xAutoRepeatOff: dpy:display -> unit = "ml_XAutoRepeatOff"
external xAutoRepeatOn: dpy:display -> unit = "ml_XAutoRepeatOn"
external xQueryKeymap: dpy:display -> string = "ml_XQueryKeymap"

external xQueryPointer: dpy:display -> win:window ->
                        window * window option * int * int * int option * int option * logical_state
    = "ml_XQueryPointer"


#if defined(MLI)
  module Got : sig
    type auto_repeat_mode = AutoRepeatModeOff | AutoRepeatModeOn
  end
  module Set : sig
    type auto_repeat_mode = AutoRepeatModeOff | AutoRepeatModeOn | AutoRepeatModeDefault
  end
#else
  module Got = struct
    type auto_repeat_mode = AutoRepeatModeOff | AutoRepeatModeOn
  end
  module Set = struct
    type auto_repeat_mode = AutoRepeatModeOff | AutoRepeatModeOn | AutoRepeatModeDefault
  end
#endif

type xKeyboardState = {
    key_click_percent: int;
    bell_percent: int;
    bell_pitch: uint;
    bell_duration: uint;
    led_mask: uint;
    global_auto_repeat: Got.auto_repeat_mode;
    auto_repeats: string;
  }
external xGetKeyboardControl: dpy:display -> xKeyboardState = "ml_XGetKeyboardControl"



(** {4 ScreenSaver} *)

type screensaver_mode =
  | ScreenSaverActive
  | ScreenSaverReset

external xForceScreenSaver: dpy:display -> mode:screensaver_mode -> unit = "ml_XForceScreenSaver"

type screensaver_values = {
    timeout: int;
    interval: int;
    prefer_blanking: int;
    allow_exposures: int;
  }
external xGetScreenSaver: dpy:display -> screensaver_values = "ml_XGetScreenSaver"
external xActivateScreenSaver: dpy:display -> unit = "ml_XActivateScreenSaver"
(*
XSetScreenSaver()
XResetScreenSaver()
*)


(* * {4 Error Handler} *)
(*
    XSetErrorHandler( ErrorHandler );


int ErrorHandler( Display *dpy, XErrorEvent *event )
{   
    char buffer[BUFSIZ];
    XGetErrorText(dpy, event->error_code, buffer, BUFSIZ);
    event->request_code;
    event->minor_code;
    event->resourceid;
    event->serial;
    return 0;
} 
*)

(** {4 Cursor} *)

type cursor_shape =
  | XC_X_cursor
  | XC_arrow
  | XC_based_arrow_down
  | XC_based_arrow_up
  | XC_boat
  | XC_bogosity
  | XC_bottom_left_corner
  | XC_bottom_right_corner
  | XC_bottom_side
  | XC_bottom_tee
  | XC_box_spiral
  | XC_center_ptr
  | XC_circle
  | XC_clock
  | XC_coffee_mug
  | XC_cross
  | XC_cross_reverse
  | XC_crosshair
  | XC_diamond_cross
  | XC_dot
  | XC_dotbox
  | XC_double_arrow
  | XC_draft_large
  | XC_draft_small
  | XC_draped_box
  | XC_exchange
  | XC_fleur
  | XC_gobbler
  | XC_gumby
  | XC_hand1
  | XC_hand2
  | XC_heart
  | XC_icon
  | XC_iron_cross
  | XC_left_ptr
  | XC_left_side
  | XC_left_tee
  | XC_leftbutton
  | XC_ll_angle
  | XC_lr_angle
  | XC_man
  | XC_middlebutton
  | XC_mouse
  | XC_pencil
  | XC_pirate
  | XC_plus
  | XC_question_arrow
  | XC_right_ptr
  | XC_right_side
  | XC_right_tee
  | XC_rightbutton
  | XC_rtl_logo
  | XC_sailboat
  | XC_sb_down_arrow
  | XC_sb_h_double_arrow
  | XC_sb_left_arrow
  | XC_sb_right_arrow
  | XC_sb_up_arrow
  | XC_sb_v_double_arrow
  | XC_shuttle
  | XC_sizing
  | XC_spider
  | XC_spraycan
  | XC_star
  | XC_target
  | XC_tcross
  | XC_top_left_arrow
  | XC_top_left_corner
  | XC_top_right_corner
  | XC_top_side
  | XC_top_tee
  | XC_trek
  | XC_ul_angle
  | XC_umbrella
  | XC_ur_angle
  | XC_watch
  | XC_xterm

external xCreateFontCursor: dpy:display -> shape:cursor_shape -> cursor = "ml_XCreateFontCursor"
external xDefineCursor: dpy:display -> win:window -> cursor:cursor -> unit = "ml_XDefineCursor"



(** {4 Window Manager} *)

external xReparentWindow: dpy:display -> win:window -> parent:window -> x:int -> y:int -> unit = "ml_XReparentWindow"

type change_mode =
  | SetModeInsert
  | SetModeDelete

external xChangeSaveSet: dpy:display -> win:window -> mode:change_mode -> unit = "ml_XChangeSaveSet"
external xAddToSaveSet: dpy:display -> win:window -> unit = "ml_XAddToSaveSet"
external xRemoveFromSaveSet: dpy:display -> win:window -> unit = "ml_XRemoveFromSaveSet"
external xKillClient: dpy:display -> resource:window -> unit = "ml_XKillClient"

(** {5 Threads} *)

external xInitThreads: unit -> unit = "ml_XInitThreads"
external xLockDisplay: dpy:display -> unit = "ml_XLockDisplay"
external xUnlockDisplay: dpy:display -> unit = "ml_XUnlockDisplay"


(** {3 ICCCM routines} *)

(*
Status XReconfigureWMWindow(
    Display*            /* display */,
    Window              /* w */,
    int                 /* screen_number */,
    unsigned int        /* mask */,
    XWindowChanges*     /* changes */
);  
    
Status XGetWMProtocols(
    Display*            /* display */,
    Window              /* w */,
    Atom**              /* protocols_return */,
    int*                /* count_return */
);  
Status XSetWMProtocols(
    Display*            /* display */,
    Window              /* w */,
    Atom*               /* protocols */,
    int                 /* count */
);
*)

external xIconifyWindow: dpy:display -> win:window -> screen_number -> unit = "ml_XIconifyWindow"
external xWithdrawWindow: dpy:display -> win:window -> screen_number -> unit = "ml_XWithdrawWindow"
external xGetCommand: dpy:display -> win:window -> string array = "ml_XGetCommand"


(*
Status XGetWMColormapWindows(
    Display*            /* display */,
    Window              /* w */,
    Window**            /* windows_return */,
    int*                /* count_return */
);
Status XSetWMColormapWindows(
    Display*            /* display */,
    Window              /* w */,
    Window*             /* colormap_windows */,
    int                 /* count */
);
void XFreeStringList(
    char**              /* list */
);
int XSetTransientForHint(
    Display*            /* display */,
    Window              /* w */,
    Window              /* prop_window */
);
*)

(* vim: sw=2 sts=2 ts=2 et fdm=marker
 *)
